# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from more_itertools import first

from orchestrator.db import ProcessStepTable, ProcessTable, SubscriptionTable
from orchestrator.utils.get_updated_properties import get_dict_updates
from orchestrator.workflow import ProcessStat, Step, StepStatus
from pydantic_forms.core import generate_form


def format_subscription(subscription: SubscriptionTable) -> dict:
    prod = subscription.product

    return {
        "subscription_id": subscription.subscription_id,
        "customer_id": subscription.customer_id,
        "product_id": prod.product_id,
        "description": subscription.description,
        "insync": subscription.insync,
        "status": subscription.status,
        "note": subscription.note,
        "start_date": subscription.start_date if subscription.start_date else None,
        "end_date": subscription.end_date if subscription.end_date else None,
        "product": {
            "product_id": prod.product_id,
            "description": prod.description,
            "name": prod.name,
            "tag": prod.tag,
            "status": prod.status,
            "product_type": prod.product_type,
        },
    }


step_finish_statuses = [StepStatus.SUCCESS, StepStatus.SKIPPED, StepStatus.COMPLETE]


def find_previous_step(steps: list[ProcessStepTable], index: int) -> ProcessStepTable | None:
    return first([step for step in reversed(steps[:index]) if step.status in step_finish_statuses], None)


def enrich_step_details(step: ProcessStepTable, previous_step: ProcessStepTable | None) -> dict:
    state_delta = get_dict_updates(previous_step.state, step.state) if previous_step else step.state

    return {
        "name": step.name,
        "executed": step.executed_at.timestamp(),
        "status": step.status,
        "state": step.state,
        "created_by": step.created_by,
        "step_id": step.step_id,
        "stepid": step.step_id,
        "state_delta": state_delta,
    }


def enrich_process_details(process: ProcessTable, p_stat: ProcessStat) -> dict:
    current_state = p_stat.state.unwrap() if p_stat.state else None
    steps: list[ProcessStepTable] = process.steps
    dict_steps = [enrich_step_details(step, find_previous_step(steps, index)) for index, step in enumerate(steps)]

    def step_fn(step: Step) -> dict:
        return {"name": step.name, "status": "pending"}

    dict_steps += list(map(step_fn, p_stat.log)) if p_stat.log else []
    form = p_stat.log[0].form if p_stat.log else None
    generated_form = generate_form(form, current_state, []) if form and current_state else None

    return {
        "steps": dict_steps,
        "form": generated_form,
        "current_state": current_state,
    }


def enrich_process(process: ProcessTable, p_stat: ProcessStat | None = None) -> dict:
    # process.subscriptions is a non JSON serializable AssociationProxy
    # So we need to build a list of Subscriptions here.
    subscriptions = [format_subscription(sub) for sub in process.subscriptions]

    details = enrich_process_details(process, p_stat) if p_stat else {}

    return {
        "process_id": process.process_id,
        "product_id": subscriptions[0]["product"]["product_id"] if subscriptions else None,
        "customer_id": subscriptions[0]["customer_id"] if subscriptions else None,
        "assignee": process.assignee,
        "last_status": process.last_status,
        "last_step": process.last_step,
        "is_task": process.is_task,
        "workflow_id": process.workflow_id,
        "workflow_name": process.workflow.name,
        "workflow_target": process.process_subscriptions[0].workflow_target if process.process_subscriptions else None,
        "failed_reason": process.failed_reason,
        "created_by": process.created_by,
        "started_at": process.started_at,
        "traceback": process.traceback,
        "last_modified_at": process.last_modified_at,
        "product": subscriptions[0]["product"] if subscriptions else None,
        "subscriptions": subscriptions,
        "steps": None,
        "form": None,
        "current_state": None,
    } | details



from http import HTTPStatus
from ipaddress import IPv4Address
from os import getenv
from unittest import mock
from uuid import uuid4

import pytest
from redis.client import Redis

from nwastdlib.url import URL
from orchestrator import app_settings
from orchestrator.api.helpers import product_block_paths
from orchestrator.db import (
    FixedInputTable,
    ProcessSubscriptionTable,
    ProcessTable,
    ProductBlockTable,
    ProductTable,
    ResourceTypeTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.db.models import SubscriptionInstanceRelationTable, WorkflowTable
from orchestrator.domain.base import SubscriptionModel
from orchestrator.services.subscriptions import (
    RELATION_RESOURCE_TYPES,
    _generate_etag,
    build_extended_domain_model,
    get_subscription,
    unsync,
)
from orchestrator.targets import Target
from orchestrator.utils.json import json_dumps, json_loads
from orchestrator.utils.redis import to_redis
from orchestrator.workflow import ProcessStatus
from test.unit_tests.config import (
    IMS_CIRCUIT_ID,
    INTERNETPINNEN_PREFIX_SUBSCRIPTION_ID,
    IPAM_PREFIX_ID,
    PARENT_IP_PREFIX_SUBSCRIPTION_ID,
    PEER_GROUP_SUBSCRIPTION_ID,
    PORT_SUBSCRIPTION_ID,
)
from test.unit_tests.conftest import do_refresh_subscriptions_search_view

SERVICE_SUBSCRIPTION_ID = str(uuid4())
PORT_A_SUBSCRIPTION_ID = str(uuid4())
PORT_A_SUBSCRIPTION_BLOCK_ID = str(uuid4())
PROVISIONING_PORT_A_SUBSCRIPTION_ID = str(uuid4())
SSP_SUBSCRIPTION_ID = str(uuid4())
SSP_SUBSCRIPTION_BLOCK_ID = str(uuid4())
IP_PREFIX_SUBSCRIPTION_ID = str(uuid4())
INVALID_SUBSCRIPTION_ID = str(uuid4())
INVALID_PORT_SUBSCRIPTION_ID = str(uuid4())

PRODUCT_ID = str(uuid4())
CUSTOMER_ID = str(uuid4())


@pytest.fixture
def seed():
    # These resource types are special
    resources = [
        ResourceTypeTable(resource_type=IMS_CIRCUIT_ID, description="Desc"),
        ResourceTypeTable(resource_type=PORT_SUBSCRIPTION_ID, description="Desc"),
    ]
    product_blocks = [
        ProductBlockTable(name="ProductBlockA", description="description a", status="active", resource_types=resources)
    ]
    fixed_inputs = [FixedInputTable(name="product_type", value="MSP100G")]
    # Wf needs to exist in code and needs to be a legacy wf with mapping

    lp_product = ProductTable(
        name="LightpathProduct",
        description="Service product that lives on ports",
        product_type="LightPath",
        tag="LightPath",  # This is special since in_use_by/depends_on subscription code handles specific tags
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    port_a_product = ProductTable(
        product_id=PRODUCT_ID,
        name="PortAProduct",
        description="Port A description",
        product_type="Port",
        tag="SP",  # This is special since in_use_by/depends_on subscription code handles specific tags
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    port_b_product = ProductTable(
        name="PortBProduct",
        description="Port B description",
        product_type="Port",
        tag="PORTB",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )

    # Special resource type handled by get subscription by ipam prefix endpoint
    ip_prefix_resources = [ResourceTypeTable(resource_type=IPAM_PREFIX_ID, description="Prefix id")]
    ip_prefix_product_blocks = [
        ProductBlockTable(
            name="ProductBlockB", description="description b", status="active", resource_types=ip_prefix_resources
        )
    ]
    ip_prefix_product = ProductTable(
        name="IPPrefixProduct",
        description="ProductTable that is used by service product",
        product_type="IP_PREFIX",
        tag="IP_PREFIX",
        status="active",
        product_blocks=ip_prefix_product_blocks,
    )
    ip_prefix_subscription = SubscriptionTable(
        subscription_id=IP_PREFIX_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=ip_prefix_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=ip_prefix_product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=ip_prefix_resources[0], value="26")],
            )
        ],
    )

    port_a_subscription = SubscriptionTable(
        subscription_id=PORT_A_SUBSCRIPTION_ID,
        description="desc",
        status="initial",
        insync=True,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    provisioning_port_a_subscription = SubscriptionTable(
        subscription_id=PROVISIONING_PORT_A_SUBSCRIPTION_ID,
        description="desc",
        status="provisioning",
        insync=False,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="12345")],
            )
        ],
    )

    ssp_subscription = SubscriptionTable(
        subscription_id=SSP_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=port_b_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    lp_subscription_instance_values_ssp = [
        SubscriptionInstanceValueTable(resource_type=resources[0], value="54321"),
        SubscriptionInstanceValueTable(resource_type=resources[1], value=str(PORT_A_SUBSCRIPTION_ID)),
    ]
    lp_subscription_instance_ssp = SubscriptionInstanceTable(
        product_block=product_blocks[0], values=lp_subscription_instance_values_ssp
    )
    lp_subscription_instance_values_msp = [
        SubscriptionInstanceValueTable(resource_type=resources[0], value="54321"),
        SubscriptionInstanceValueTable(resource_type=resources[1], value=str(SSP_SUBSCRIPTION_ID)),
    ]
    lp_subscription_instance_msp = SubscriptionInstanceTable(
        product_block=product_blocks[0], values=lp_subscription_instance_values_msp
    )
    lp_subscription = SubscriptionTable(
        subscription_id=SERVICE_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=lp_product,
        customer_id=CUSTOMER_ID,
        instances=[lp_subscription_instance_ssp, lp_subscription_instance_msp],
    )

    invalid_subscription = SubscriptionTable(
        subscription_id=INVALID_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[],
    )

    invalid_tagged_product = ProductTable(
        product_id=str(uuid4()),
        name="INVALID_PRODUCT",
        description="invalid descr",
        product_type="Port",
        tag="NEWMSP",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )

    invalid_tagged_subscription = SubscriptionTable(
        subscription_id=INVALID_PORT_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=False,
        product=invalid_tagged_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )
    db.session.add(port_a_product)
    db.session.add(ip_prefix_product)
    db.session.add(ip_prefix_subscription)
    db.session.add(invalid_tagged_product)
    db.session.add(port_b_product)
    db.session.add(lp_product)
    db.session.add(port_a_subscription)
    db.session.add(provisioning_port_a_subscription)
    db.session.add(ssp_subscription)
    db.session.add(lp_subscription)
    db.session.add(invalid_subscription)
    db.session.add(invalid_tagged_subscription)
    db.session.commit()

    RELATION_RESOURCE_TYPES.extend(
        [
            PORT_SUBSCRIPTION_ID,
            IP_PREFIX_SUBSCRIPTION_ID,
            INTERNETPINNEN_PREFIX_SUBSCRIPTION_ID,
            PARENT_IP_PREFIX_SUBSCRIPTION_ID,
            PEER_GROUP_SUBSCRIPTION_ID,
        ]
    )

    return ip_prefix_product


# seeder to test direct relations using SubscriptionInstanceRelationTable instead of resource type.
@pytest.fixture
def seed_with_direct_relations():
    # These resource types are special
    resources = [
        ResourceTypeTable(resource_type=IMS_CIRCUIT_ID, description="Desc"),
    ]
    product_blocks = [
        ProductBlockTable(name="ProductBlockA", description="description a", status="active", resource_types=resources)
    ]
    fixed_inputs = [FixedInputTable(name="product_type", value="MSP100G")]
    # Wf needs to exist in code and needs to be a legacy wf with mapping

    lp_product = ProductTable(
        name="LightpathProduct",
        description="Service product that lives on ports",
        product_type="LightPath",
        tag="LightPath",  # This is special since in_use_by/depends_on subscription code handles specific tags
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    port_a_product = ProductTable(
        product_id=PRODUCT_ID,
        name="PortAProduct",
        description="Port A description",
        product_type="Port",
        tag="SP",  # This is special since in_use_by/depends_on subscription code handles specific tags
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    port_b_product = ProductTable(
        name="PortBProduct",
        description="Port B description",
        product_type="Port",
        tag="PORTB",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )

    # Special resource type handled by get subscription by ipam prefix endpoint
    ip_prefix_resources = [ResourceTypeTable(resource_type=IPAM_PREFIX_ID, description="Prefix id")]
    ip_prefix_product_blocks = [
        ProductBlockTable(
            name="ProductBlockB", description="description b", status="active", resource_types=ip_prefix_resources
        )
    ]
    ip_prefix_product = ProductTable(
        name="IPPrefixProduct",
        description="ProductTable that is used by service product",
        product_type="IP_PREFIX",
        tag="IP_PREFIX",
        status="active",
        product_blocks=ip_prefix_product_blocks,
    )
    ip_prefix_subscription = SubscriptionTable(
        subscription_id=IP_PREFIX_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=ip_prefix_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=ip_prefix_product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=ip_prefix_resources[0], value="26")],
            )
        ],
    )

    port_a_subscription = SubscriptionTable(
        subscription_id=PORT_A_SUBSCRIPTION_ID,
        description="desc",
        status="initial",
        insync=True,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                subscription_instance_id=PORT_A_SUBSCRIPTION_BLOCK_ID,
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    provisioning_port_a_subscription = SubscriptionTable(
        subscription_id=PROVISIONING_PORT_A_SUBSCRIPTION_ID,
        description="desc",
        status="provisioning",
        insync=False,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="12345")],
            )
        ],
    )

    ssp_subscription = SubscriptionTable(
        subscription_id=SSP_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=port_b_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                subscription_instance_id=SSP_SUBSCRIPTION_BLOCK_ID,
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    lp_subscription_instance_ssp_id = str(uuid4())
    lp_subscription_instance_values_ssp = [
        SubscriptionInstanceValueTable(resource_type=resources[0], value="54321"),
    ]
    lp_subscription_instance_ssp = SubscriptionInstanceTable(
        subscription_instance_id=lp_subscription_instance_ssp_id,
        product_block=product_blocks[0],
        values=lp_subscription_instance_values_ssp,
    )
    lp_subscription_instance_ssp_depends_on = SubscriptionInstanceRelationTable(
        in_use_by_id=lp_subscription_instance_ssp_id,
        depends_on_id=PORT_A_SUBSCRIPTION_BLOCK_ID,
        order_id=0,
        domain_model_attr="service_port",
    )

    lp_subscription_instance_msp_id = str(uuid4())
    lp_subscription_instance_values_msp = [
        SubscriptionInstanceValueTable(resource_type=resources[0], value="54321"),
    ]
    lp_subscription_instance_msp = SubscriptionInstanceTable(
        subscription_instance_id=lp_subscription_instance_msp_id,
        product_block=product_blocks[0],
        values=lp_subscription_instance_values_msp,
    )
    lp_subscription_instance_msp_depends_on = SubscriptionInstanceRelationTable(
        in_use_by_id=lp_subscription_instance_msp_id,
        depends_on_id=SSP_SUBSCRIPTION_BLOCK_ID,
        order_id=0,
        domain_model_attr="service_port",
    )
    lp_subscription = SubscriptionTable(
        subscription_id=SERVICE_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=lp_product,
        customer_id=CUSTOMER_ID,
        instances=[lp_subscription_instance_ssp, lp_subscription_instance_msp],
    )

    invalid_subscription = SubscriptionTable(
        subscription_id=INVALID_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[],
    )

    invalid_tagged_product = ProductTable(
        product_id=str(uuid4()),
        name="INVALID_PRODUCT",
        description="invalid descr",
        product_type="Port",
        tag="NEWMSP",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )

    invalid_tagged_subscription = SubscriptionTable(
        subscription_id=INVALID_PORT_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=False,
        product=invalid_tagged_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    db.session.add(port_a_product)
    db.session.add(ip_prefix_product)
    db.session.add(ip_prefix_subscription)
    db.session.add(invalid_tagged_product)
    db.session.add(port_b_product)
    db.session.add(lp_product)
    db.session.add(port_a_subscription)
    db.session.add(provisioning_port_a_subscription)
    db.session.add(ssp_subscription)
    db.session.add(lp_subscription)
    db.session.add(invalid_subscription)
    db.session.add(invalid_tagged_subscription)
    db.session.commit()
    db.session.add(lp_subscription_instance_ssp_depends_on)
    db.session.add(lp_subscription_instance_msp_depends_on)
    db.session.commit()

    RELATION_RESOURCE_TYPES.extend(
        [
            PORT_SUBSCRIPTION_ID,
            IP_PREFIX_SUBSCRIPTION_ID,
            INTERNETPINNEN_PREFIX_SUBSCRIPTION_ID,
            PARENT_IP_PREFIX_SUBSCRIPTION_ID,
            PEER_GROUP_SUBSCRIPTION_ID,
        ]
    )

    return ip_prefix_product


def test_insync_404(seed, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{str(uuid4())}")
    assert response.status_code == HTTPStatus.NOT_FOUND


def test_insync_invalid_tagged(seed, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{INVALID_PORT_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    assert response.json() == {
        "reason": "subscription.not_in_sync",
        "create": [],
        "modify": [],
        "terminate": [],
        "system": [],
    }


def test_in_use_by_subscriptions_not_insync(seed, test_client):
    # ensure that the used subscription of the MSP is out of sync
    service = get_subscription(SERVICE_SUBSCRIPTION_ID)
    service.insync = False
    db.session.commit()

    # test the api endpoint
    response = test_client.get(f"/api/subscriptions/workflows/{PORT_A_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    insync_info = response.json()
    assert "reason" in insync_info
    assert len(insync_info["locked_relations"]) == 1
    assert insync_info["locked_relations"][0] == SERVICE_SUBSCRIPTION_ID


def test_in_use_by_subscriptions_insync(seed, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{PORT_A_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK

    insync_info = response.json()
    assert "reason" not in insync_info
    assert "locked_relations" not in insync_info


def test_depends_on_subscriptions_not_insync(seed, test_client):
    # ensure that the depends on subscription of the LP is out of sync
    msp = db.session.get(SubscriptionTable, PORT_A_SUBSCRIPTION_ID, with_for_update=True)
    msp.insync = False
    db.session.commit()

    # test the api endpoint
    response = test_client.get(f"/api/subscriptions/workflows/{SERVICE_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    insync_info = response.json()
    assert "reason" in insync_info
    assert len(insync_info["locked_relations"]) == 1
    assert insync_info["locked_relations"][0] == PORT_A_SUBSCRIPTION_ID


def test_depends_on_subscriptions_insync(seed, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{SERVICE_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK

    insync_info = response.json()
    assert "reason" not in insync_info
    assert "locked_relations" not in insync_info


def test_in_use_by_subscriptions_not_insync_direct_relations(seed_with_direct_relations, test_client):
    # ensure that the used subscription of the MSP is out of sync
    service = get_subscription(SERVICE_SUBSCRIPTION_ID)
    service.insync = False
    db.session.commit()

    # test the api endpoint
    response = test_client.get(f"/api/subscriptions/workflows/{PORT_A_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    insync_info = response.json()
    assert "reason" in insync_info
    assert len(insync_info["locked_relations"]) == 1
    assert insync_info["locked_relations"][0] == SERVICE_SUBSCRIPTION_ID


def test_depends_on_subscriptions_not_insync_direct_relations(seed_with_direct_relations, test_client):
    # ensure that the depends on subscription of the LP is out of sync
    msp = get_subscription(PORT_A_SUBSCRIPTION_ID, for_update=True)
    msp.insync = False
    db.session.commit()

    # test the api endpoint
    response = test_client.get(f"/api/subscriptions/workflows/{SERVICE_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    insync_info = response.json()
    assert "reason" in insync_info
    assert len(insync_info["locked_relations"]) == 1
    assert insync_info["locked_relations"][0] == PORT_A_SUBSCRIPTION_ID


def test_depends_on_subscriptions_insync_direct_relations(seed_with_direct_relations, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{SERVICE_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK

    insync_info = response.json()
    assert "reason" not in insync_info
    assert "locked_relations" not in insync_info


def test_delete_subscription_404(responses, seed, test_client):
    sub_id = uuid4()

    response = test_client.delete(f"/api/subscriptions/{sub_id}")
    assert response.status_code == HTTPStatus.NOT_FOUND


def test_set_in_sync(seed, test_client):
    subscription_id = IP_PREFIX_SUBSCRIPTION_ID
    unsync(subscription_id)
    db.session.commit()

    response = test_client.put(f"/api/subscriptions/{subscription_id}/set_in_sync")
    assert response.status_code == HTTPStatus.OK

    subscription = get_subscription(subscription_id)
    assert subscription.insync


def _create_failed_process(subscription_id):
    wf = WorkflowTable(workflow_id=uuid4(), name="validate_ip_prefix", target=Target.SYSTEM)
    process_id = uuid4()
    process = ProcessTable(
        process_id=process_id,
        workflow_id=wf.workflow_id,
        last_status=ProcessStatus.FAILED,
        last_step="Verify references in NSO",
        assignee="NOC",
        is_task=False,
    )
    process_subscription = ProcessSubscriptionTable(process_id=process_id, subscription_id=subscription_id)
    db.session.add(wf)
    db.session.add(process)
    db.session.add(process_subscription)

    db.session.commit()


def test_try_set_failed_task_in_sync(seed, test_client):
    subscription_id = IP_PREFIX_SUBSCRIPTION_ID
    unsync(IP_PREFIX_SUBSCRIPTION_ID)
    db.session.commit()

    _create_failed_process(subscription_id)

    response = test_client.put(f"/api/subscriptions/{subscription_id}/set_in_sync")
    assert response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY

    subscription = get_subscription(subscription_id)
    assert not subscription.insync


def test_product_block_paths(generic_subscription_1, generic_subscription_2):
    subscription_1 = SubscriptionModel.from_subscription(generic_subscription_1)
    subscription_2 = SubscriptionModel.from_subscription(generic_subscription_2)
    assert product_block_paths(subscription_1) == ["product", "pb_1", "pb_2"]
    assert product_block_paths(subscription_2) == ["product", "pb_3"]


@pytest.mark.parametrize(
    "query, num_matches",
    [
        ("id", 7),
        ("tag:(POR* | LP)", 1),
        ("tag:SP", 3),
        ("tag:(POR* | LP) | tag:SP", 4),
        ("tag:(POR* | LP) | tag:SP -status:initial", 3),
    ],
)
def test_subscriptions_search(query, num_matches, seed, test_client, refresh_subscriptions_search_view):
    response = test_client.get(f"/api/subscriptions/search?query={query}")
    result = response.json()
    assert len(result) == num_matches


@pytest.fixture
def make_port_subscription():

    def make(subscription_id=None, customer_id=None, product_id=None):
        product_id = str(product_id or uuid4())
        subscription_id = str(subscription_id or uuid4())
        customer_id = str(customer_id or uuid4())
        product = ProductTable(
            product_id=product_id,
            name="ProductTable",
            description="description",
            product_type="Port",
            tag="Port",
            status="active",
        )
        subscription = SubscriptionTable(
            subscription_id=subscription_id,
            description="desc",
            status="active",
            insync=True,
            product=product,
            customer_id=customer_id,
        )

        db.session.add(product)
        db.session.add(subscription)
        db.session.commit()
        do_refresh_subscriptions_search_view()
        return subscription_id, customer_id, product_id

    return make


@pytest.mark.parametrize(
    "subscription_id",
    [
        "144cd85c-77a9-43bf-a78c-3ef2647c6ff1",
        "67bb12ae-8dfc-4ddb-9fd2-979392a9f2ad",
        "85ab4e94-c045-46f3-85c7-a8fbc7411aca",
        "27e84752-c700-491e-b126-454b5fa06dc0",
        "191b9852-e689-4dd2-8699-fc9c11c2bcdd",
        "998e7c12-e4d2-4321-a075-04b13a76ab39",
        "cfc64927-8f52-4325-87ac-daa1e16794e8",
        "5846383e-6387-411b-8e6e-bca022e2a86d",
    ],
)
def test_subscriptions_search_uuids2(
    subscription_id, make_port_subscription, test_client, refresh_subscriptions_search_view
):
    make_port_subscription(subscription_id=subscription_id)

    search_queries = {"full": subscription_id} | {
        f"group{n}": part for n, part in enumerate(subscription_id.split("-"))
    }

    def search(keyword):
        response = test_client.get(f"/api/subscriptions/search?query={keyword}")
        return len(response.json()) == 1

    results = {testcase: search(keyword) for testcase, keyword in search_queries.items()}

    succeeded = [search_queries[testcase] for testcase in results if results[testcase]]
    failed = [search_queries[testcase] for testcase in results if not results[testcase]]

    assert not failed, f"Could not find '{subscription_id}' by all keywords; {succeeded=} {failed=}"


def test_subscription_detail_with_domain_model(test_client, generic_subscription_1):
    # test with a subscription that has domain model and without
    response = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response.status_code == HTTPStatus.OK
    # Check hierarchy
    assert response.json()["pb_1"]["rt_1"] == "Value1"


def test_subscription_detail_with_domain_model_does_not_exist(test_client, generic_subscription_1):
    # test with a subscription that has domain model and without
    response = test_client.get(URL("api/subscriptions/domain-model") / uuid4())
    assert response.status_code == HTTPStatus.NOT_FOUND


def test_subscription_detail_with_domain_model_etag(test_client, generic_subscription_1):
    # test with a subscription that has domain model and without
    response = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response.status_code == HTTPStatus.OK
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)
    etag = _generate_etag(extended_model)
    assert etag == response.headers["ETag"]
    # Check hierarchy
    assert response.json()["pb_1"]["rt_1"] == "Value1"


def test_subscription_detail_with_domain_model_if_none_match(test_client, generic_subscription_1):
    # test with a subscription that has domain model and without
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)
    etag = _generate_etag(extended_model)
    response = test_client.get(
        URL("api/subscriptions/domain-model") / generic_subscription_1, headers={"If-None-Match": etag}
    )
    assert response.status_code == HTTPStatus.NOT_MODIFIED


@pytest.mark.skipif(
    not getenv("AIOCACHE_DISABLE", "0") == "0", reason="AIOCACHE must be enabled for this test to do anything"
)
def test_subscription_detail_with_domain_model_cache(test_client, generic_subscription_1):
    # test with a subscription that has domain model and without
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)
    etag = _generate_etag(extended_model)

    app_settings.CACHE_DOMAIN_MODELS = True

    to_redis(extended_model)

    response = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)

    cache = Redis.from_url(str(app_settings.CACHE_URI))
    result = cache.get(f"orchestrator:domain:{generic_subscription_1}")
    cached_model = json_dumps(json_loads(result))
    cached_etag = cache.get(f"orchestrator:domain:etag:{generic_subscription_1}")
    assert cached_model == json_dumps(extended_model)
    assert cached_etag.decode("utf-8") == etag

    assert response.status_code == HTTPStatus.OK
    assert response.json()["subscription_id"] == generic_subscription_1
    app_settings.CACHE_DOMAIN_MODELS = False
    cache.delete(f"orchestrator:domain:{generic_subscription_1}")


def test_subscription_detail_with_in_use_by_ids_filtered_self(test_client, product_one_subscription_1):
    response = test_client.get(URL("api/subscriptions/domain-model") / product_one_subscription_1)
    assert response.status_code == HTTPStatus.OK
    assert not response.json()["block"]["sub_block"]["in_use_by_ids"]


@mock.patch("orchestrator.api.api_v1.endpoints.subscriptions.get_subscription_dict")
def test_subscription_detail_special_fields(mock_from_redis, test_client):
    """Test that a subscription with special field types is correctly serialized by Pydantic.

    https://github.com/pydantic/pydantic/issues/6669
    """
    standard_fields = {
        "subscription_id": "fabd6359-cb37-4a1c-bfc4-5c15aea7c888",
        "description": "desc",
        "status": "active",
        "customer_id": "f711c6fe-6de3-40bd-a4e7-9ac9d183a788",
        "insync": True,
        "product": {
            "name": "fake name",
            "description": "fake description",
            "product_type": "fake type",
            "status": "active",
            "tag": "fake tag",
        },
    }
    # Make the from_redis function return an IPv4Address - this wouldn't happen normally but it's easier
    # to mock than SubscriptionModel.from_subscription, and tests the special field formatting all the same
    special_fields = {"ip_address": IPv4Address("127.0.0.1")}
    mock_from_redis.return_value = (standard_fields | special_fields, "etag ofzo")

    response = test_client.get(URL("api/subscriptions/domain-model") / "fabd6359-cb37-4a1c-bfc4-5c15aea7c888")
    assert response.json() == {
        "subscription_id": "fabd6359-cb37-4a1c-bfc4-5c15aea7c888",
        "start_date": None,
        "description": "desc",
        "status": "active",
        "product_id": None,
        "customer_id": "f711c6fe-6de3-40bd-a4e7-9ac9d183a788",
        "insync": True,
        "note": None,
        "name": None,
        "end_date": None,
        "product": {
            "product_id": None,
            "name": "fake name",
            "description": "fake description",
            "product_type": "fake type",
            "status": "active",
            "tag": "fake tag",
            "created_at": None,
            "end_date": None,
        },
        "customer_descriptions": [],
        "tag": None,
        "ip_address": "127.0.0.1",
    }


def test_subscription_detail_with_in_use_by_ids_not_filtered_self(test_client, product_one_subscription_1):
    response = test_client.get(
        URL("api/subscriptions/domain-model") / product_one_subscription_1 / "?filter_owner_relations=false"
    )
    assert response.status_code == HTTPStatus.OK
    assert response.json()["block"]["sub_block"]["in_use_by_ids"]



from uuid import UUID

from orchestrator.domain.base import SubscriptionModel
from orchestrator.services.subscriptions import _generate_etag, build_extended_domain_model
from orchestrator.utils.redis import from_redis


async def get_subscription_dict(subscription_id: UUID) -> tuple[dict, str]:
    """Helper function to get subscription dict by uuid from db or cache."""

    if cached_model := from_redis(subscription_id):
        return cached_model  # type: ignore

    subscription_model = SubscriptionModel.from_subscription(subscription_id)
    subscription = build_extended_domain_model(subscription_model)
    etag = _generate_etag(subscription)
    return subscription, etag



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that provides service functions on subscriptions."""
import pickle  # noqa: S403
from collections import defaultdict
from collections.abc import Sequence
from datetime import datetime
from hashlib import md5
from typing import Any, TypeVar, overload
from uuid import UUID

import more_itertools
import structlog
from sqlalchemy import Text, cast, not_, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Query, aliased, joinedload
from sqlalchemy.sql.expression import or_

from orchestrator.api.helpers import getattr_in, product_block_paths, update_in
from orchestrator.db import (
    ProductTable,
    ResourceTypeTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.db.models import (
    SubscriptionCustomerDescriptionTable,
    SubscriptionInstanceRelationTable,
    SubscriptionMetadataTable,
)
from orchestrator.domain.base import SubscriptionModel
from orchestrator.targets import Target
from orchestrator.types import SubscriptionLifecycle, UUIDstr
from orchestrator.utils.datetime import nowtz
from orchestrator.utils.helpers import is_ipaddress_type

logger = structlog.get_logger(__name__)

T = TypeVar("T", bound=SubscriptionTable)


@overload
def get_subscription(subscription_id: UUID | UUIDstr, for_update: bool = False) -> SubscriptionTable: ...


@overload
def get_subscription(
    subscription_id: UUID | UUIDstr, for_update: bool = False, model: type[T] = SubscriptionTable
) -> T: ...


def get_subscription(
    subscription_id: UUID | UUIDstr, for_update: bool = False, model: type[T] = SubscriptionTable
) -> T:
    """Get the subscription.

    Args:
        subscription_id: The subscription_id
        for_update: specific whether we intend to update the subscription
        model: SubscriptionModelType

    Returns: A subscription object

    Raises: ValueError: if the requested Subscription does not exist in de database.

    """

    try:
        subscription = db.session.get(model, subscription_id, with_for_update=for_update)
    except SQLAlchemyError as e:
        raise ValueError("Invalid subscription id") from e

    if subscription:
        return subscription
    raise ValueError(f"Subscription with {subscription_id} does not exist in the database")


def get_subscription_metadata(subscription_id: UUIDstr) -> dict | None:
    subscription_metadata = SubscriptionMetadataTable.find_by_subscription_id(subscription_id)
    if subscription_metadata:
        return subscription_metadata.metadata_
    return None


def update_subscription_status(subscription_id: UUIDstr, status: str) -> SubscriptionTable:
    """Update the subscription status.

    Args:
        subscription_id: Id of the subscription to update.
        status: Status to transition to.

    Returns: Subscription

    """
    subscription = get_subscription(subscription_id, for_update=True)
    subscription.status = status
    return subscription


def update_subscription_description(subscription_id: UUIDstr, description: str) -> SubscriptionTable:
    """Update a subscription's description.

    Args:
        subscription_id: subscription id of the subscription to update
        description: new subscription description

    Returns: Subscription

    """
    subscription = get_subscription(subscription_id, for_update=True)
    subscription.description = description
    return subscription


def activate_subscription(subscription_id: UUIDstr) -> SubscriptionTable:
    """Activate subscription by subscription id.

    Args:
        subscription_id: subscription id of the subscription

    Returns: Subscription object

    """
    subscription = get_subscription(subscription_id, for_update=True)
    subscription.status = "active"
    subscription.start_date = nowtz()
    subscription.insync = True
    return subscription


def provision_subscription(subscription_id: UUIDstr) -> SubscriptionTable:
    """Provision subscription by subscription id.

    Args:
        subscription_id: Subscription id of the subscription

    Returns: Updated subscription object

    """
    subscription = get_subscription(subscription_id, for_update=True)
    subscription.status = "provisioning"
    subscription.insync = True
    return subscription


def migrate_subscription(subscription_id: UUIDstr) -> SubscriptionTable:
    """Migrate subscription by subscription id.

    Args:
        subscription_id: Subscription id of the subscription

    Returns: Updated subscription object

    """
    subscription = get_subscription(subscription_id, for_update=True)
    subscription.status = "migrating"
    subscription.insync = False
    return subscription


def unsync(subscription_id: UUIDstr, checked: bool = True) -> SubscriptionTable:
    """Unsync subscription by subscription id.

    Args:
        subscription_id: Subscription id of the subscription
        checked: Checked or not

    Returns: Updated subscription object

    """
    subscription = get_subscription(subscription_id, for_update=True)
    if checked and not subscription.insync:
        raise ValueError("Subscription is already out of sync, cannot continue!")
    subscription.insync = False
    return subscription


def resync(subscription_id: UUIDstr) -> SubscriptionTable:
    """Resync subscription by subscription id.

    Args:
        subscription_id: Subscription id of the subscription

    Returns: Updated subscription object

    """
    subscription = get_subscription(subscription_id, for_update=True)
    subscription.insync = True
    return subscription


def set_status_provisioning_subscription(subscription_id: UUIDstr) -> SubscriptionTable:
    return update_subscription_status(subscription_id, "provisioning")


def terminate_subscription(subscription_id: UUIDstr) -> SubscriptionTable:
    return update_subscription(subscription_id, status="terminated", end_date=nowtz())


def create_subscription(
    customer_id: str, product: ProductTable, subscription_name: str, subscription_id: UUIDstr
) -> UUID:
    subscription = SubscriptionTable(
        subscription_id=subscription_id,
        product_id=product.product_id,
        customer_id=customer_id,
        description=subscription_name,
        start_date=None,
        end_date=None,
        insync=False,
        status="initial",
    )
    db.session.add(subscription)

    return subscription.subscription_id


def update_subscription(subscription_id: str, **attrs: dict | UUIDstr | str | datetime) -> SubscriptionTable:
    """Update the subscription.

    Args:
        subscription_id: SubscriptionTable id of the subscription
        attrs: Attributes that will be set

    Returns: Subscription

    """

    subscription = get_subscription(subscription_id, for_update=True)

    for key, value in attrs.items():
        setattr(subscription, key, value)

    return subscription


def retrieve_node_subscriptions_by_name(node_name: str) -> list[SubscriptionTable]:
    stmt = (
        select(SubscriptionTable)
        .join(ProductTable)
        .join(SubscriptionInstanceTable)
        .join(SubscriptionInstanceValueTable)
        .join(ResourceTypeTable)
        .filter(SubscriptionInstanceValueTable.value == node_name)
        .filter(ResourceTypeTable.resource_type == "nso_device_id")
        .filter(SubscriptionTable.status.in_(["active", "provisioning"]))
    )
    return list(db.session.scalars(stmt))


def retrieve_subscription_by_subscription_instance_value(
    resource_type: str, value: str, sub_status: tuple = ("provisioning", "active")
) -> SubscriptionTable | None:
    """Retrieve a Subscriptions by resource_type and value.

    Args:
        resource_type: name of the resource type
        value: value of the resource type
        sub_status: status of the subscriptions

    Returns: Subscription or None

    """
    stmt = (
        select(SubscriptionTable)
        .join(SubscriptionInstanceTable)
        .join(SubscriptionInstanceValueTable)
        .join(ResourceTypeTable)
        .filter(SubscriptionInstanceValueTable.value == value)
        .filter(ResourceTypeTable.resource_type == resource_type)
        .filter(SubscriptionTable.status.in_(sub_status))
        .distinct(SubscriptionTable.subscription_id)
    )
    return db.session.scalars(stmt).one_or_none()


def find_values_for_resource_types(
    subscription_id: UUID | UUIDstr, resource_types: Sequence[str], strict: bool = True
) -> dict[str, list[str]]:
    """Find values for resource types by subscription ID.

    This function issues a single SQL query to find one or more resource types for a given subscription. As
    multiple values per resource type are possible (think of a BGP subscription with multiple SAPs, with each SAP
    sharing the same resource types) values are always returned as a list, even when there is only a single value.

    In case of shared resource types (hence multiple values), the order of the values matches across resource types:
    Meaning: in case of the above example with multiple, say two, SAPs, should we have requested the resource types::

        ('customer_ipv4_mtu', 'customer_ipv6_mtu', 'port_subscription_id'),

    we would get back::

        {
            'customer_ipv4_mtu': ['1500', '9000'],
            'customer_ipv6_mtu': ['9000', '9000'],
            'port_subscription_id': ['654a1e7f-9afb-43b7-ba8f-cdf9d48494aa', '2f86be06-a5b7-40ee-a71b-137d6b48a37e']
        }

    The first element of each list belong to the same SAP, likewise the second element of each list belongs to the
    same SAP. This allows you to implicitly correlate values.

    Args:
        subscription_id: The id of the subscription.
        resource_types: A sequence of resource type names.
        strict:
            True: raise `ValueError` if one or more requested resource types were not found,
            False: ignore resource types that weren't found.

    Returns:
        A dictionary of resource type names to lists of values.

    Raises:
        ValueError: if strict == True and one or more resource types were requested but not found.

    """
    # the `order_by` on `subscription_instance_id` is there to guarantee the matched ordering across resource_types
    # (see also docstring)
    stmt = (
        select(SubscriptionInstanceValueTable)
        .join(ResourceTypeTable)
        .join(SubscriptionInstanceTable)
        .filter(
            SubscriptionInstanceTable.subscription_id == subscription_id,
            ResourceTypeTable.resource_type.in_(resource_types),
        )
        .order_by(SubscriptionInstanceTable.subscription_instance_id)
        .with_only_columns(ResourceTypeTable.resource_type, SubscriptionInstanceValueTable.value)
    )
    resource_type_values = db.session.execute(stmt).all()

    rt2v: dict[str, list[str]] = defaultdict(list)
    for resource_type, value in resource_type_values:
        rt2v[resource_type].append(value)
    if strict:
        missing = set(resource_types) - set(rt2v.keys())
        if missing:
            raise ValueError(f"Could not find requested resource types: '{','.join(missing)}'!")
    return rt2v


def query_in_use_by_subscriptions(subscription_id: UUID, filter_statuses: list[str] | None = None) -> Query:
    """Return a query with all subscriptions -in_use_by- that use this subscription with resource_type or direct relation.

    The query can be used to add extra filters when/where needed.
    """
    # Find relations through resource types
    resource_type_relations = (
        SubscriptionTable.query.join(SubscriptionInstanceTable)
        .options(joinedload(SubscriptionTable.customer_descriptions))
        .join(SubscriptionInstanceValueTable)
        .join(ResourceTypeTable)
        .filter(ResourceTypeTable.resource_type.in_(RELATION_RESOURCE_TYPES))
        .filter(SubscriptionInstanceValueTable.value == str(subscription_id))
        .with_entities(SubscriptionTable.subscription_id)
    )

    # Find relations through instance hierarchy
    in_use_by_instances = aliased(SubscriptionInstanceTable)
    depends_on_instances = aliased(SubscriptionInstanceTable)
    relation_relations = (
        SubscriptionTable.query.join(in_use_by_instances.subscription)
        .join(in_use_by_instances.depends_on_block_relations)
        .join(depends_on_instances, SubscriptionInstanceRelationTable.depends_on)
        .filter(depends_on_instances.subscription_id == subscription_id)
        .filter(in_use_by_instances.subscription_id != subscription_id)
        .with_entities(SubscriptionTable.subscription_id)
    )

    return SubscriptionTable.query.filter(
        or_(
            SubscriptionTable.subscription_id.in_(resource_type_relations.scalar_subquery()),
            SubscriptionTable.subscription_id.in_(relation_relations.scalar_subquery()),
        ),
        SubscriptionTable.status.in_(filter_statuses if filter_statuses else SubscriptionLifecycle.values()),
    )


def query_depends_on_subscriptions(subscription_id: UUID, filter_statuses: list[str] | None = None) -> Query:
    """Return a query with all subscriptions -depends_on- that this subscription is dependent on with resource_type or direct relation.

    The query can be used to add extra filters when/where needed.
    """
    # Find relations through resource types
    resource_type_relations = (
        SubscriptionInstanceTable.query.join(SubscriptionInstanceValueTable)
        .join(ResourceTypeTable)
        .filter(ResourceTypeTable.resource_type.in_(RELATION_RESOURCE_TYPES))
        .filter(SubscriptionInstanceTable.subscription_id == subscription_id)
        .join(SubscriptionTable, SubscriptionInstanceValueTable.value == cast(SubscriptionTable.subscription_id, Text))
        .with_entities(SubscriptionTable.subscription_id)
    )

    # Find relations through instance hierarchy
    in_use_by_instances = aliased(SubscriptionInstanceTable)
    depends_on_instances = aliased(SubscriptionInstanceTable)
    relation_relations = (
        SubscriptionTable.query.join(depends_on_instances.subscription)
        .join(depends_on_instances.in_use_by_block_relations)
        .join(in_use_by_instances, SubscriptionInstanceRelationTable.in_use_by)
        .filter(in_use_by_instances.subscription_id == subscription_id)
        .filter(depends_on_instances.subscription_id != subscription_id)
        .with_entities(SubscriptionTable.subscription_id)
    )

    return SubscriptionTable.query.filter(
        or_(
            SubscriptionTable.subscription_id.in_(resource_type_relations.scalar_subquery()),
            SubscriptionTable.subscription_id.in_(relation_relations.scalar_subquery()),
        ),
        SubscriptionTable.status.in_(filter_statuses if filter_statuses else SubscriptionLifecycle.values()),
    )


def _terminated_filter(query: Query) -> list[UUID]:
    return list(
        more_itertools.flatten(
            query.filter(SubscriptionTable.status != "terminated").with_entities(SubscriptionTable.subscription_id)
        )
    )


def _in_sync_filter(query: Query) -> list[UUID]:
    return list(
        more_itertools.flatten(
            query.filter(not_(SubscriptionTable.insync)).with_entities(SubscriptionTable.subscription_id)
        )
    )


RELATION_RESOURCE_TYPES: list[str] = []


def status_relations(subscription: SubscriptionTable | None) -> dict[str, list[UUID]]:
    """Return info about locked subscription dependencies.

    This call will be used by the client to determine if it's safe to
    start a MODIFY or TERMINATE workflow. There are 4 cases:

    1) The subscription is a IP, LightPath or ELAN: the depends_on subscriptions are checked for not 'insync' instances.
    2) The subscription is a ServicePort: in_use_by subscriptions are checked for not 'insync' instances and for in_use_by
       services that are not terminated.
    3) The subscription is a Node: Related Core link subscriptions are checked that there are no active instances
       This is only used for the terminate workflow and ignored for modify
    4) IP_prefix cannot be terminated when in use

    """
    if not subscription:
        return {"locked_relations": [], "unterminated_parents": [], "unterminated_in_use_by_subscriptions": []}
    in_use_by_query = query_in_use_by_subscriptions(subscription.subscription_id)

    unterminated_in_use_by_subscriptions = _terminated_filter(in_use_by_query)
    locked_in_use_by_block_relations = _in_sync_filter(in_use_by_query)

    depends_on_query = query_depends_on_subscriptions(subscription.subscription_id)

    locked_depends_on_block_relations = _in_sync_filter(depends_on_query)

    result = {
        "locked_relations": locked_in_use_by_block_relations + locked_depends_on_block_relations,
        "unterminated_parents": unterminated_in_use_by_subscriptions,
        "unterminated_in_use_by_subscriptions": unterminated_in_use_by_subscriptions,
    }

    logger.debug(
        "Returning status info for related subscriptions",
        result=result,
        subscription_id=str(subscription.subscription_id),
    )
    return result


def get_relations(subscription_id: UUIDstr) -> dict[str, list[UUID]]:
    subscription_table = db.session.get(
        SubscriptionTable,
        subscription_id,
        options=[
            joinedload(SubscriptionTable.product),
            joinedload(SubscriptionTable.product).joinedload(ProductTable.workflows),
        ],
    )
    return status_relations(subscription_table)


TARGET_DEFAULT_USABLE_MAP: dict[Target, list[str]] = {
    Target.CREATE: [],
    Target.MODIFY: ["active"],
    Target.TERMINATE: ["active", "provisioning"],
    Target.SYSTEM: ["active"],
}

WF_USABLE_MAP: dict[str, list[str]] = {}

WF_BLOCKED_BY_PARENTS: dict[str, bool] = {}
WF_BLOCKED_BY_IN_USE_BY_SUBSCRIPTIONS: dict[str, bool] = {}

WF_USABLE_WHILE_OUT_OF_SYNC: list[str] = ["modify_note"]


def subscription_workflows(subscription: SubscriptionTable) -> dict[str, Any]:
    """Return a dict containing all the workflows a user can start for this subscription.

    Args:
        subscription: an SqlAlchemy instance of a `db.SubscriptionTable`

    Returns:
        A dictionary with the following structure (reason and its related keys are only present when workflows are blocked):

        >>> {  # doctest:+SKIP
        ...     "reason": "Optional global reason like subscription is in use"
        ...     "create": [{"name": "workflow.name", "description": "workflow.description", "reason": "Optional reason why this specific workflow is blocked"}],
        ...     "modify": [],
        ...     "terminate": [],
        ...     "system": [],
        ... }

    """
    default_json: dict[str, Any] = {}

    if not subscription.insync:
        default_json["reason"] = "subscription.not_in_sync"
    else:
        data = status_relations(subscription)

        if data["locked_relations"]:
            default_json["reason"] = "subscription.relations_not_in_sync"
            default_json["locked_relations"] = data["locked_relations"]

    workflows: dict[str, Any] = {
        "create": [],
        "modify": [],
        "terminate": [],
        "system": [],
    }
    for workflow in subscription.product.workflows:
        if workflow.name in WF_USABLE_WHILE_OUT_OF_SYNC or workflow.target == Target.SYSTEM:
            # validations and modify note are also possible with: not in sync or locked relations
            workflow_json = {"name": workflow.name, "description": workflow.description}
        else:
            workflow_json = {"name": workflow.name, "description": workflow.description, **default_json}

        # Simple reasons like "not in sync" and "locked_relations" are handled now. Handle special lifecycle cases:
        if "reason" not in workflow_json:
            default = TARGET_DEFAULT_USABLE_MAP[workflow.target]
            usable_when = WF_USABLE_MAP.get(workflow.name, default)

            if subscription.status not in usable_when:
                workflow_json["reason"] = "subscription.no_modify_invalid_status"
                workflow_json["usable_when"] = usable_when
                workflow_json["status"] = subscription.status
                workflow_json["action"] = "terminated" if workflow.target == Target.TERMINATE else "modified"

            # Check if this workflow is blocked because there are unterminated relations
            blocked_by_depends_on_subscriptions = WF_BLOCKED_BY_IN_USE_BY_SUBSCRIPTIONS.get(
                workflow.name, workflow.target == Target.TERMINATE
            )

            if not blocked_by_depends_on_subscriptions:
                blocked_by_depends_on_subscriptions = WF_BLOCKED_BY_PARENTS.get(
                    workflow.name, workflow.target == Target.TERMINATE
                )
            if blocked_by_depends_on_subscriptions and data["unterminated_in_use_by_subscriptions"]:
                workflow_json["reason"] = "subscription.no_modify_subscription_in_use_by_others"
                workflow_json["unterminated_parents"] = data["unterminated_parents"]
                workflow_json["unterminated_in_use_by_subscriptions"] = data["unterminated_in_use_by_subscriptions"]
                workflow_json["action"] = "terminated" if workflow.target == Target.TERMINATE else "modified"

        workflows[workflow.target.lower()].append(workflow_json)

    return {**workflows, **default_json}


def _generate_etag(model: dict) -> str:
    encoded = pickle.dumps(model)
    return md5(encoded).hexdigest()  # noqa: S303, S324


def convert_to_in_use_by_relation(obj: Any) -> dict[str, str]:
    return {"subscription_instance_id": str(obj.subscription_instance_id), "subscription_id": str(obj.subscription_id)}


def build_extended_domain_model(subscription_model: SubscriptionModel) -> dict:
    """Create a subscription dict from the SubscriptionModel with additional keys."""
    stmt = select(SubscriptionCustomerDescriptionTable).filter(
        SubscriptionCustomerDescriptionTable.subscription_id == subscription_model.subscription_id
    )
    customer_descriptions = list(db.session.scalars(stmt))

    subscription = subscription_model.model_dump()
    paths = product_block_paths(subscription)

    def inject_in_use_by_ids(path_to_block: str) -> None:
        if not (in_use_by_subs := getattr_in(subscription_model, f"{path_to_block}.in_use_by")):
            return

        in_use_by_ids = [obj.in_use_by_id for obj in in_use_by_subs.col]
        in_use_by_relations = [convert_to_in_use_by_relation(instance) for instance in in_use_by_subs]
        update_in(subscription, f"{path_to_block}.in_use_by_ids", in_use_by_ids)
        update_in(subscription, f"{path_to_block}.in_use_by_relations", in_use_by_relations)

    # find all product blocks, check if they have in_use_by and inject the in_use_by_ids into the subscription dict.
    for path in paths:
        inject_in_use_by_ids(path)

    subscription["customer_descriptions"] = customer_descriptions

    return subscription


def format_special_types(subscription: dict) -> dict:
    """Modifies the subscription dict in-place, formatting special types to string.

    This function was added during the Pydantic 2.x migration to handle serialization errors on ipaddress types.
    Background: https://github.com/pydantic/pydantic/issues/6669

    The problem lies with SubscriptionDomainModelSchema which allows extra untyped fields.
    It might be possible with a model_serializer but couldn't get this to work, therefore this workaround.
    """

    def format_value(v: Any) -> Any:
        if is_ipaddress_type(v):
            return str(v)
        if isinstance(v, dict):
            return format_special_types(v)
        if isinstance(v, list):
            return [format_value(item) for item in v]
        return v

    for k, v in subscription.items():
        subscription[k] = format_value(v)
    return subscription


def format_extended_domain_model(subscription: dict, filter_owner_relations: bool) -> dict:
    """Format the subscription dict depending on filter settings.

    Args:
        subscription: result from build_extended_domain_model() or cache
        filter_owner_relations: True to filter instance ids from the current subscription
    """

    def filter_instance_ids_on_subscription() -> None:
        paths = product_block_paths(subscription)
        instance_ids_to_filter = {getattr_in(subscription, f"{path}.subscription_instance_id") for path in paths}
        instance_ids_to_filter.add(subscription["subscription_id"])

        def filter_instance_ids_on_productblock(path_to_block: str) -> None:
            if not (block_instance_ids := getattr_in(subscription, f"{path_to_block}.in_use_by_ids")):
                return
            filtered_instance_ids = set(block_instance_ids) - instance_ids_to_filter
            update_in(subscription, f"{path_to_block}.in_use_by_ids", list(filtered_instance_ids))

        # find all product blocks, filter instance_ids if needed
        for path in paths:
            filter_instance_ids_on_productblock(path)

    if filter_owner_relations:
        filter_instance_ids_on_subscription()

    return subscription



from unittest.mock import patch

from dirty_equals import IsFloat, IsInt

from orchestrator.db.listeners import disable_listeners, monitor_sqlalchemy_queries
from orchestrator.settings import app_settings


def test_stats_extension(fastapi_app_graphql, test_client):
    # given
    query = """query MyQuery {
      workflows(first: 10) {
        page {
          name
        }
      }
    }"""
    try:
        monitor_sqlalchemy_queries()
        with patch.object(app_settings, "ENABLE_GRAPHQL_STATS_EXTENSION", True):
            fastapi_app_graphql.register_graphql()

            # when
            response = test_client.post("/api/graphql", json={"query": query})
    finally:
        disable_listeners()

    # then
    result = response.json()
    assert result["extensions"]["stats"] == {"db_queries": IsInt, "db_time": IsFloat, "operation_time": IsFloat}
